from cubes.datacat import cwsource_pull_data
from cubes.datacat.vocabularies import add_source

sync_schema_props_perms('Dataset')
sync_schema_props_perms('dataset_contact_point')
sync_schema_props_perms('dataset_publisher')
sync_schema_props_perms('license_type')
sync_schema_props_perms('license')

# Since in_catalog in now mandatory, put orphan datasets into a default catalog
rset = rql('Agent A WHERE A name "admin"')  # Ensure there is a default publisher
if not rset:
    admin = cnx.create_entity('Agent', name=u'admin')
else:
    admin = rset.one()
cat = cnx.create_entity('DataCatalog', title=u'System Catalog',  # Create the default catalog
                        description=u'Default catalog for orphan datasets',
                        catalog_publisher=admin)
commit(ask_confirm=False)
rql('SET DSET in_catalog CAT WHERE NOT EXISTS(DSET in_catalog C), CAT eid %(cat_eid)s',
    {'cat_eid': cat.eid})
commit(ask_confirm=False)

# French translations
print('In order to have french translations, please synchronize again using '
      '"cubicweb-ctl sync-schemes <instance>" command')

# Synchronize licenses
print('-> creating SKOS source for licenses and importing licenses')
lic_skos_source = add_source(cnx, u'European Licences Named Authority List',
                             u'http://publications.europa.eu/mdr/resource/authority/'
                             u'licence/skos/licences-skos.rdf')
commit(ask_confirm=False)
cwsource_pull_data(repo, lic_skos_source.through_cw_source[0].eid)
print('')

add_relation_type('scheme_relation')
# Licenses are now concepts: convert LicenseDocument entities to Concept entities
lic_scheme = cnx.find('ConceptScheme',
                      cwuri=u'http://publications.europa.eu/resource/authority/licence').one()
rql('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name "license_type"',
    {'cs': lic_scheme.eid})
commit(ask_confirm=False)
uri2concept = dict(  # Build some mappings to match existing concepts
    rql('(Any U,C WHERE C in_scheme S, S eid %(scheme_eid)s, C cwuri U) UNION '
        '(Any U,C2 WHERE C2 in_scheme S, S eid %(scheme_eid)s, C2 exact_match C, C cwuri U)',
        {'scheme_eid': lic_scheme.eid})
)
title2concept = dict(
    rql('Any LL,C WHERE C in_scheme S, S eid %(scheme_eid)s, L label_of C, L label LL',
        {'scheme_eid': lic_scheme.eid})
)
add_relation_definition('DataCatalog', 'license', 'Concept')  # Actual migration
add_relation_definition('Distribution', 'license', 'Concept')
add_relation_definition('Concept', 'license_type', 'Concept')
for license in cnx.find('LicenseDocument').entities():
    # Look for existing concept: match URI first, then title
    lic_concept_eid = uri2concept.get(license.uri) or title2concept.get(license.title)
    if lic_concept_eid:
        rql('SET X license C WHERE X license L, C is Concept, L is LicenseDocument, '
            'C eid %(concept_eid)s, L eid %(license_eid)s',
            {'concept_eid': lic_concept_eid, 'license_eid': license.eid})
    else:  # No existing concept found, create a new one
        kwargs = {}
        if license.uri:
            kwargs['cwuri'] = license.uri
        if license.license_type:
            kwargs['license_type'] = license.license_type[0]
        lic_concept = cnx.create_entity('Concept', in_scheme=lic_scheme, **kwargs)
        cnx.create_entity('Label', label_of=lic_concept, label=license.title,
                          kind=u"preferred",
                          language_code=unicode(cnx.vreg.property_value('ui.language')))
        commit(ask_confirm=False)
        rql('SET X license C WHERE C is Concept, C eid %(concept_eid)s',
            {'concept_eid': lic_concept.eid})
        commit(ask_confirm=False)
drop_entity_type('LicenseDocument')



from cubicweb import NoResultError

add_attribute('Dataset', 'ckan_organization')
# Move ckan_organization from Application to Dataset.
rql('SET X ckan_organization O '
    'WHERE X is Dataset, X application A, A ckan_organization O')
commit()
# Convert Application application_maintainer to Dataset contact_point.
rset = rql('Any X,U WHERE X application A, A application_maintainer U')
for dataset, user in zip(rset.entities(0), rset.entities(1)):
    dataset.cw_set(contact_point=user.dc_title())
commit()
drop_entity_type('Application')
sync_schema_props_perms('Dataset')

add_attribute('Dataset', 'landing_page')
add_attribute('Dataset', 'frequency')
add_attribute('Dataset', 'spatial_coverage')
add_attribute('Dataset', 'provenance')
add_attribute('Dataset', 'theme')

add_attribute('Agent', 'email')
sync_schema_props_perms('Agent')
add_relation_definition('Dataset', 'dataset_contact_point', 'Agent')
add_relation_definition('Dataset', 'dataset_contributors', 'Agent')

rset = rql('Any X,A WHERE NOT X contact_point NULL, X contact_point A')
for eid, name in rset.rows:
    try:
        cpoint = find_one_entity('Agent', name=name)
    except NoResultError:
        _ = create_entity('Agent', name=name,
                          reverse_dataset_contact_point=eid)
    else:
        cpoint.cw_set(reverse_dataset_contact_point=eid)
commit()
drop_attribute('Dataset', 'contact_point')

drop_attribute('Agent', 'publisher_type')



from os.path import join, dirname


process_script(join(dirname(__file__), 'import_dcat_vocabularies.py'))

drop_attribute('Dataset', 'theme')
drop_attribute('Agent', 'dcat_type')
add_relation_definition('Agent', 'publisher_type', 'Concept')
drop_attribute('Dataset', 'frequency')
add_relation_definition('Dataset', 'dataset_frequency', 'Concept')

drop_attribute('DataCatalog', 'license')
drop_attribute('Distribution', 'licence')
add_entity_type('LicenseDocument')



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat tests for geo-spatial views."""

import json

from cubicweb.devtools import (
    PostgresApptestConfiguration,
    startpgcluster,
    stoppgcluster,
    testlib,
)
from cubicweb.web import facet as cwfacet

from cubes.datacat.views.geo import (
    ExtentFacet,
    RelationExtentFacet,
)


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class ExtentFacetTC(testlib.CubicWebTC):
    """Test case for the geographical extent facet."""

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        polygon = (
            u'MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),(2 2,2 3,3 3,3 2,2 2)),((6 3,9 2,9 4,6 3)))',
            2154,
        )
        with self.admin_access.cnx() as cnx:
            thing = cnx.create_entity(
                'Thing', title=u'thing',
                polygon=polygon,
            )
            cnx.create_entity(
                'OtherThing', name=u'other',
                thing_related=thing,
                geomcoll=(u'GEOMETRYCOLLECTION (POINT (2 46))', 2154),
            )
            cnx.commit()

    def _new_facet(self, req, facet_cls, initial_rql, mainvar, rtype, role=None, target_attr=None):
        """Return a new facet with the given parameters.

        ``req`` is the web request using the facet.

        The new facet will belong to the ``facet_cls`` class.

        The new facet will be initialized with the ``initial_rql`` RQL query in which ``mainvar`` is
        the main variable, and with ``rtype``, ``role``, and ``target_attr`` class attributes.
        """
        rset = req.cnx.execute(initial_rql)
        rqlst = rset.syntax_tree().copy()
        filtered_variable, baserql = cwfacet.init_facets(rset, rqlst.children[0], mainvar=mainvar)
        facet = facet_cls(req, rset=rset, select=rqlst.children[0],
                          filtered_variable=filtered_variable)
        facet.__regid__ = 'extent_facet'
        facet.rtype = rtype
        if role is not None:
            facet.role = role
        if target_attr is not None:
            facet.target_attr = target_attr
        return facet

    def test_no_widget_if_no_geometry(self):
        with self.admin_access.cnx() as cnx:
            catalog = cnx.create_entity('DataCatalog', title=u'cat')
            cnx.create_entity('Dataset', in_catalog=catalog)
            cnx.commit()
        with self.admin_access.web_request() as req:
            facet = self._new_facet(req,
                                    facet_cls=ExtentFacet,
                                    initial_rql='Location X',
                                    mainvar='X',
                                    rtype='geometry')
            self.assertIsNone(facet.get_widget())
            facet = self._new_facet(req,
                                    facet_cls=RelationExtentFacet,
                                    initial_rql='Dataset X',
                                    mainvar='X',
                                    rtype='locate_datasets',
                                    role='object',
                                    target_attr='geometry')
            self.assertIsNone(facet.get_widget())

    def test_convex_hull_relation(self):
        """Check that widget is displayed with convex hull geospatial data (relation case)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=RelationExtentFacet,
                                    initial_rql='OtherThing X',
                                    mainvar='X',
                                    rtype='thing_related',
                                    role='subject',
                                    target_attr='polygon')
            # Check convex hull
            widget = facet.get_widget()
            rset = widget.rset
            self.assertEqual(rset.rql,
                             u'DISTINCT Any ST_ASGEOJSON(ST_CONVEXHULL(ST_COLLECT(B))) '
                             u'WHERE X is OtherThing, X thing_related A, A polygon B '
                             u'HAVING ST_COLLECT(B) != ""')
            expected = {
                u'type': u'Polygon',
                u'coordinates': [[[1, 1], [1, 5], [5, 5],
                                  [9, 4], [9, 2], [5, 1], [1, 1]]],
            }
            self.assertEqual(len(rset), 1)
            self.assertEqual(json.loads(rset[0][0]), expected)

    def test_convex_hull(self):
        """Check that widget is displayed with convex hull geospatial data (no relation case)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=ExtentFacet,
                                    initial_rql='Thing X',
                                    mainvar='X',
                                    rtype='polygon')
            # Check convex hull
            widget = facet.get_widget()
            rset = widget.rset
            self.assertEqual(rset.rql,
                             u'DISTINCT Any ST_ASGEOJSON(ST_CONVEXHULL(ST_COLLECT(A))) '
                             u'WHERE X is Thing, X polygon A HAVING ST_COLLECT(A) != ""')
            expected = {
                u'type': u'Polygon',
                u'coordinates': [[[1, 1], [1, 5], [5, 5],
                                  [9, 4], [9, 2], [5, 1], [1, 1]]],
            }
            self.assertEqual(len(rset), 1)
            self.assertEqual(json.loads(rset[0][0]), expected)

    def test_rql_restriction_relation(self):
        """Check RQL restriction when a geometry is drawn in the facet (relation case)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=RelationExtentFacet,
                                    initial_rql='OtherThing X',
                                    mainvar='X',
                                    rtype='thing_related',
                                    role='subject',
                                    target_attr='polygon')
            self.assertEqual(facet.select.as_string(), 'DISTINCT Any  WHERE X is OtherThing')
            # Nothing drawn -> no new restriction
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(), 'DISTINCT Any  WHERE X is OtherThing')
            # Geometry drawn if the facet
            facet._cw.form[facet.__regid__ + 'Input'] = '<geojson_geom_str>'
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(),
                             u'DISTINCT Any  WHERE X is OtherThing, X thing_related A, '
                             u'A polygon B HAVING ST_INTERSECTS('
                             u'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             u'B) = TRUE')

    def test_rql_restriction(self):
        """Check RQL restriction when a geometry is drawn in the facet (no relation case)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=ExtentFacet,
                                    initial_rql='Thing X',
                                    mainvar='X',
                                    rtype='polygon')
            self.assertEqual(facet.select.as_string(), 'DISTINCT Any  WHERE X is Thing')
            # Nothing drawn -> no new restriction
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(), 'DISTINCT Any  WHERE X is Thing')
            # Geometry drawn if the facet
            facet._cw.form[facet.__regid__ + 'Input'] = '<geojson_geom_str>'
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(),
                             u'DISTINCT Any  WHERE X is Thing, X polygon A '
                             u'HAVING ST_INTERSECTS('
                             u'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             u'A) = TRUE')

    def test_rql_restriction_existing_having(self):
        """Check RQL restriction when a geometry is drawn in the facet (existing having clause)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=ExtentFacet,
                                    initial_rql='Thing X WHERE X title T HAVING T = "My Title"',
                                    mainvar='X',
                                    rtype='polygon')
            self.assertEqual(facet.select.as_string(),
                             'DISTINCT Any  WHERE X title T, X is Thing HAVING T = "My Title"')
            # Geometry drawn if the facet
            facet._cw.form[facet.__regid__ + 'Input'] = '<geojson_geom_str>'
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(),
                             u'DISTINCT Any  WHERE X title T, X is Thing, X polygon A '
                             u'HAVING ST_INTERSECTS('
                             u'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             u'A) = TRUE, T = "My Title"')

    def test_rql_restriction_geometry_collection(self):
        """Check RQL restriction when a geometry is drawn in the facet (geometry collection
        case)."""
        with self.admin_access.web_request() as req:
            # Create facet
            facet = self._new_facet(req,
                                    facet_cls=ExtentFacet,
                                    initial_rql='OtherThing X',
                                    mainvar='X',
                                    rtype='geomcoll')
            self.assertEqual(facet.select.as_string(), 'DISTINCT Any  WHERE X is OtherThing')
            # Geometry drawn if the facet
            facet._cw.form[facet.__regid__ + 'Input'] = '<geojson_geom_str>'
            facet.add_rql_restrictions()
            self.assertEqual(facet.select.as_string(),
                             'DISTINCT Any  WHERE X is OtherThing, X geomcoll A HAVING '
                             '((ST_INTERSECTS('
                             'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             'ST_COLLECTIONEXTRACT(A, 1)) = TRUE) OR '
                             '(ST_INTERSECTS('
                             'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             'ST_COLLECTIONEXTRACT(A, 2)) = TRUE)) OR '
                             '(ST_INTERSECTS('
                             'ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON("<geojson_geom_str>"), 3), '
                             'ST_COLLECTIONEXTRACT(A, 3)) = TRUE)')


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat automatic tests"""

from cubicweb.devtools import (
    PostgresApptestConfiguration,
    startpgcluster,
    stoppgcluster,
    testlib,
)

from cubes.datacat import (
    cwsource_pull_data,
    vocabularies,
)


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class VocabularyImportTC(testlib.CubicWebTC):
    """Functional tests ensuring concept schemes got properly imported.
    """

    configcls = PostgresApptestConfiguration

    def test_adms(self):
        """Just check one import from SKOSSource"""
        scheme_names = (
            'assettype',
            'interoperabilitylevel',
            'licencetype',
            'publishertype',
            'representationtechnique',
            'status',
        )
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute(
                'Any X WHERE S through_cw_source X, S url LIKE "%ADMS%"')
            assert rset
            for eid, in rset.rows:
                cwsource_pull_data(self.repo, eid)
            rset = cnx.execute(
                'Any X WHERE X is ConceptScheme, X cwuri IN ({0})'.format(
                    ','.join('"http://purl.org/adms/{0}/1.0"'.format(n)
                             for n in scheme_names)
                )
            )
        self.assertEqual(len(rset), len(scheme_names))

    def test_mediatypes(self):
        """Check import of concepts from IANA "Media Types"."""
        with self.admin_access.repo_cnx() as cnx:
            vocabularies.media_types_import(cnx, media_types=('audio', 'image'))
            cnx.commit()
            rset = cnx.execute(
                'Any COUNT(X) WHERE X in_scheme C, C cwuri "{0}"'.format(
                    'http://www.iana.org/assignments/media-types/media-types.xml')
            )
            self.assertEqual(rset[0][0], 192)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat tests for geo-spatial entities"""

from cubicweb.devtools import (
    PostgresApptestConfiguration,
    startpgcluster,
    stoppgcluster,
    testlib,
)


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class IGeoJSONTC(testlib.CubicWebTC):
    """Test case for converting entities to IGeoJSON interface."""

    configcls = PostgresApptestConfiguration

    def test_location_as_geojson(self):
        """Check that locations are correctly converted to IGeoJSON interface."""
        with self.admin_access.repo_cnx() as cnx:
            # No geometry
            loc1 = cnx.create_entity('Location', name=u'Location 1')
            cnx.commit()
            igeojson1 = loc1.cw_adapt_to('IGeoJSON')
            self.assertEqual(igeojson1.as_geojson(), {})
            # With geometry
            loc2 = cnx.create_entity('Location', name=u'Location 2',
                                     geometry=(u'GEOMETRYCOLLECTION (POINT (2 46))', 4326))
            cnx.commit()
            igeojson2 = loc2.cw_adapt_to('IGeoJSON')
            self.assertEqual(igeojson2.as_geojson(),
                             {'type': u'Feature', u'id': loc2.cwuri,
                              'properties': {'name': u'Location 2',
                                             'html_info': u'<strong>Location 2</strong>'},
                              'geometry': {'type': u'GeometryCollection',
                                           'geometries': [{'type': u'Point',
                                                           'coordinates': [2, 46]}]}})

    def test_dataset_as_geojson(self):
        """Check that datasets correctly return a GeoJSON object."""
        with self.admin_access.repo_cnx() as cnx:
            # No related location
            publisher = cnx.create_entity('Agent', name=u'publisher')
            catalog = cnx.create_entity('DataCatalog', title=u'cata',
                                        description=u'catalog',
                                        catalog_publisher=publisher)
            dset = cnx.create_entity('Dataset', title=u'My Dataset',
                                     description=u'desc', in_catalog=catalog)
            cnx.commit()
            self.assertIsNone(dset.cw_adapt_to('IGeoJSON'))
            # One related location with no geometry
            cnx.create_entity('Location', name=u'Location 1', locate_datasets=dset.eid)
            cnx.commit()
            dset.cw_clear_all_caches()
            igeojson = dset.cw_adapt_to('IGeoJSON')
            self.assertEqual(igeojson.as_geojson(), {})
            # One related location with geometry
            loc2 = cnx.create_entity('Location', name=u'Location 2',
                                     geometry=(u'GEOMETRYCOLLECTION (POINT (2 46))', 4326),
                                     locate_datasets=dset.eid)
            cnx.commit()
            igeojson2 = loc2.cw_adapt_to('IGeoJSON')
            self.assertEqual(igeojson.as_geojson(),
                             {'type': 'FeatureCollection', 'features': [igeojson2.as_geojson()]})


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat geo-spatial utilities."""


#
# Geometry helper functions
#


def srs_from_epsg(epsg):
    """Returns an ``osr.SpatialReference`` instance from the given EPSG code.

    Raises a ``ValueError`` if ``epsg`` is not a valid EPSG code.
    """
    from osgeo import osr
    srs = osr.SpatialReference()
    try:
        srs.ImportFromEPSG(int(epsg))
    except (TypeError, ValueError):
        raise ValueError('Invalid EPSG: {}'.format(epsg))
    return srs


def ogr_geom(value):
    """Returns an ``ogr.Geometry`` instance from the given value.

    Raises a ``ValueError`` if no geometry can be created from the given value.

    This function accepts values in the following formats:

    * (x, y) (2-uple of integers),
    * WKT,
    * GeoJSON.

    """
    from osgeo import ogr
    ogr.UseExceptions()  # See https://trac.osgeo.org/gdal/wiki/PythonGotchas
    if isinstance(value, (tuple, list)):  # Value has (x, y) format
        x, y = value
        geom = ogr.Geometry(ogr.wkbPoint)
        geom.SetCoordinateDimension(2)
        geom.AddPoint_2D(x, y)
        return geom
    for geom_function in [ogr.CreateGeometryFromWkt,  # Try in turn WKT, GeoJSON
                          ogr.CreateGeometryFromJson]:
        try:
            return geom_function(value)
        except RuntimeError:
            pass
    raise ValueError('Invalid value for Geometry: {}'.format(value))


# Checker function
# Needed until cubicweb-postgis does so by itself (see https://www.cubicweb.org/ticket/5629998)

def is_geometry(value, epsg=None, expected_geom_type=None, expected_epsg=None):
    """Returns ``True`` if the given value and srid are valid according to the expected parameters.
    """
    from osgeo import ogr
    ogr.UseExceptions()  # See https://trac.osgeo.org/gdal/wiki/PythonGotchas
    # Consider epsg
    if epsg is not None and not isinstance(epsg, int):
        return False
    if expected_epsg is not None and (epsg is None or epsg != expected_epsg):
        return False
    # Consider value
    try:  # Value must be WKT
        geom = ogr.CreateGeometryFromWkt(value)
        geom.IsValid()
    except RuntimeError:
        return False
    if expected_geom_type is not None \
            and geom.GetGeometryName().upper() != expected_geom_type.upper():
        return False
    return True


# Converter function
# Needed until cubicweb-postgis does so by itself (see https://www.cubicweb.org/ticket/5629998)

def geometry(value, epsg=None, expected_geom_type=None, expected_epsg=None):
    """Converts the given value according to the specified parameters, and returns a 2-uple
    ``(wkt, srid)`` suitable for cubicweb-postgis.

    The returned ``wkt`` is a string representing a geometry value in Well-Known Text (WKT) syntax,
    while ``srid`` is an integer identifying the spatial reference system in which lies ``wkt``.

    If the spatial reference system cannot be determined, returned ``srid`` is 0.

    Raises a ``ValueError`` if the given value cannot be converted to a valid ``Geometry``.

    Examples:

    .. code-block:: python

        >>> geometry('POINT (2 46)', expected_geom_type='MultiPoint')
        ('MULTIPOINT ((2 46))', 4326)

    .. code-block:: python

        >>> geometry('POINT (2 46)', expected_geom_type='GeometryCollection')
        ('GEOMETRYCOLLECTION (POINT (2 46))', 0)

    .. code-block:: python

        >>> geometry('POINT (2 46)', epsg=4326, expected_epsg=2154)
        ('POINT (622609.119992160820402 6544963.910583730787039)', 2154))

    .. code-block:: python

        >>> geometry('POINT (2 46)', expected_geom_type='LineString')
        Traceback (most recent call last):
          ...
        ValueError: Cannot convert 'POINT (2 46)' TO 'LINESTRING'

    """
    from osgeo import osr, ogr
    ogr.UseExceptions()  # See https://trac.osgeo.org/gdal/wiki/PythonGotchas
    _convert_to_multi_function = {
        'MULTIPOINT': ogr.ForceToMultiPoint,
        'MULTILINESTRING': ogr.ForceToMultiLineString,
        'MULTIPOLYGON': ogr.ForceToMultiPolygon,
    }
    # Create spatial reference systems from EPSGs
    srs = srs_from_epsg(epsg) if epsg is not None else None
    expected_srs = srs_from_epsg(expected_epsg) if expected_epsg is not None else None
    # Create geometry from value
    geom = ogr_geom(value)
    # Convert to expected type
    if expected_geom_type is not None:
        expected_geom_type = expected_geom_type.upper()
        geom_type = geom.GetGeometryName().upper()
        if geom_type == expected_geom_type:
            pass
        elif expected_geom_type.endswith(geom_type):  # Expect MULTITYPE and have TYPE
            geom = _convert_to_multi_function[expected_geom_type](geom)
        elif expected_geom_type == 'GEOMETRYCOLLECTION':
            geomc = ogr.Geometry(ogr.wkbGeometryCollection)
            geomc.AddGeometry(geom)
            geom = geomc
        else:
            raise ValueError('Cannot convert {} to {}'.format(value, expected_geom_type))
    # Reproject to expected spatial reference system
    if srs is not None:
        geom.AssignSpatialReference(srs)
    srs = geom.GetSpatialReference()  # May have been assigned elsewhere (eg. inline in GeoJSON)
    if expected_srs is not None:
        if srs is None:
            raise ValueError('Invalid or missing origin EPSG: {}'.format(epsg))
        if not srs.IsSame(expected_srs):
            transform = osr.CoordinateTransformation(srs, expected_srs)
            geom.Transform(transform)
            geom.AssignSpatialReference(expected_srs)
            srs = expected_srs
    # Return 2-uple
    srid = srs.GetAttrValue('Authority', 1) if srs is not None else None
    srid = int(srid) if srid is not None else 0
    return geom.ExportToWkt(), srid


#
# GeoJSON helper functions
#

def merge_in_geojson_collection(geojsons):
    """Return a new ``FeatureCollection`` GeoJSON object after merging all the objects in the given
    iterable.

    Return an empty object if both objects are empty too.

    Example:

    .. code-block:: python

        >>> geojson1 = {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [2, 46]}}
        >>> geosjon2 = {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [2, 47]}}
        >>> merge = merge_geojson_features([geojson1, geosjon2])
        >>> type(merge)
        dict
        >>> merge['type']
        u'FeatureCollection'
        >>> len(merge['features'])
        2
        >>> merge['features'][0]
        {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [2, 46]}}
        >>> merge['features'][1]
        {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [2, 47]}}

    As another example, merging a ``FeatureCollection`` with a ``Feature`` will append the
    ``Feature`` to list of features in the ``FeatureCollection``.
    """
    features = []
    for geojson in geojsons:
        if 'type' in geojson:
            if geojson['type'] == 'Feature':
                features.append(geojson)
            elif geojson['type'] == 'FeatureCollection':
                for feature in geojson['features']:
                    features.append(feature)
    if not features:
        return {}
    return {'type': u'FeatureCollection', 'features': features}



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-datacat vocabularies (concept schemes)"""

from os.path import (
    dirname,
    join,
)
import re

from cubicweb.dataimport import (
    massive_store,
    stores,
    ucsvreader,
)
from cubicweb.dataimport.importer import (
    ExtEntity,
    SimpleImportLog,
)

from cubes.skos import rdfio
from cubes.skos.sobjects import import_skos_extentities


DCAT_SCHEMES = [
    (u'ADMS vocabularies',
     u'https://joinup.ec.europa.eu/svn/adms/ADMS_v1.00/ADMS_SKOS_v1.00.rdf'),
    (u'European Frequency Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/frequency/skos/frequencies-skos.rdf'),
    (u'European Filetypes Authority Table',
     u'http://publications.europa.eu/mdr/resource/authority/file-type/skos/filetypes-skos.rdf'),
    (u'European Languages Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/language/skos/languages-skos.rdf'),
    (u'European Dataset Theme Vocabulary',
     u'http://publications.europa.eu/mdr/resource/authority/data-theme/skos/data-theme-skos.rdf'),
    (u'European Licences Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/licence/skos/licences-skos.rdf'),
    (u'European Country Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/country/skos/countries-skos.rdf'),
    (u'European Places Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/place/skos/places-skos.rdf'),
    (u'European Continents Named Authority List',
     u'http://publications.europa.eu/mdr/resource/authority/continent/skos/continents-skos.rdf'),
]

EUROVOC_URI = (u'http://publications.europa.eu/mdr/resource/thesaurus/eurovoc-20150416-0/skos/'
               'eurovoc_skos.zip')

LICENSE_SCHEME = u'http://publications.europa.eu/resource/authority/licence'

ADMS_FRENCH_LABELS = {
    'http://purl.org/adms/assettype/CodeList': u'Liste de codes',
    'http://purl.org/adms/assettype/CoreComponent': u'Composant commun',
    'http://purl.org/adms/assettype/DomainModel': u'Modèle de domaine',
    # 'http://purl.org/adms/assettype/InformationExchangePackageDescription': u'',
    'http://purl.org/adms/assettype/Mapping': u'Correspondance',
    'http://purl.org/adms/assettype/NameAuthorityList': u"Liste de noms d'autorité",
    'http://purl.org/adms/assettype/Ontology': u'Ontologie',
    'http://purl.org/adms/assettype/Schema': u'Schéma',
    'http://purl.org/adms/assettype/ServiceDescription': u"Description d'un service",
    'http://purl.org/adms/assettype/SyntaxEncodingScheme': u"Système d'encodage",
    'http://purl.org/adms/assettype/Taxonomy': u'Taxonomie',
    'http://purl.org/adms/assettype/Thesaurus': u'Thésaurus',
    'http://purl.org/adms/interoperabilitylevel/Legal': u'Légal',
    'http://purl.org/adms/interoperabilitylevel/Organisational': u'Organisationnel',
    'http://purl.org/adms/interoperabilitylevel/Political': u'Politique',
    'http://purl.org/adms/interoperabilitylevel/Semantic': u'Sémantique',
    'http://purl.org/adms/interoperabilitylevel/Technical': u'Technique',
    'http://purl.org/adms/licencetype/Attribution': u'Attribution',
    # 'http://purl.org/adms/licencetype/GrantBack': u'',
    'http://purl.org/adms/licencetype/JurisdictionWithinTheEU': (
        u"Compétence d'une juridiction de l'UE"
    ),
    'http://purl.org/adms/licencetype/KnownPatentEncumbrance': u'Protégé par brevet',
    'http://purl.org/adms/licencetype/NoDerivativeWork': u'Pas de modification',
    'http://purl.org/adms/licencetype/NominalCost': u'Contribution nominale',
    'http://purl.org/adms/licencetype/NonCommercialUseOnly': u"Pas d'utilisation commerciale",
    'http://purl.org/adms/licencetype/OtherRestrictiveClauses': u'Autres clauses restrictives',
    'http://purl.org/adms/licencetype/PublicDomain': u'Domaine public',
    'http://purl.org/adms/licencetype/ReservedNames-Endorsement-OfficialStatus': (
        u'Utilisation du nom ou de la marque interdite'
    ),
    'http://purl.org/adms/licencetype/RoyaltiesRequired': u'Soumis à redevance',
    'http://purl.org/adms/licencetype/UnknownIPR': u'Inconnu',
    'http://purl.org/adms/licencetype/ViralEffect-ShareAlike': u'Partage dans les mêmes conditions',
    'http://purl.org/adms/publishertype/Academia-ScientificOrganisation': (
        u'Académique/Organisation scientifique'
    ),
    'http://purl.org/adms/publishertype/Company': u'Société',
    'http://purl.org/adms/publishertype/IndustryConsortium': u'Consortium industriel',
    'http://purl.org/adms/publishertype/LocalAuthority': u'Collectivité locale',
    'http://purl.org/adms/publishertype/NationalAuthority': u'Administration nationale',
    'http://purl.org/adms/publishertype/NonGovernmentalOrganisation': (
        u'Organisation non gouvernementale'
    ),
    'http://purl.org/adms/publishertype/NonProfitOrganisation': u'Organisation à but non lucratif',
    'http://purl.org/adms/publishertype/PrivateIndividual(s)': u'Individu(s)',
    'http://purl.org/adms/publishertype/RegionalAuthority': u'Autorité régionale',
    'http://purl.org/adms/publishertype/StandardisationBody': u'Organisme de standardisation',
    'http://purl.org/adms/publishertype/SupraNationalAuthority': u'Autorité supra-nationale',
    'http://purl.org/adms/representationtechnique/Archimate': u'Archimate',
    'http://purl.org/adms/representationtechnique/BPMN': u'BPMN',
    'http://purl.org/adms/representationtechnique/CommonLogic': u'Common logic',
    'http://purl.org/adms/representationtechnique/DTD': u'DTD',
    'http://purl.org/adms/representationtechnique/Datalog': u'Datalog',
    'http://purl.org/adms/representationtechnique/Diagram': u'Diagramme',
    'http://purl.org/adms/representationtechnique/Genericode': u'genericode',
    'http://purl.org/adms/representationtechnique/HumanLanguage': u'Langue naturelle',
    'http://purl.org/adms/representationtechnique/IDEF': u'IDEF',
    'http://purl.org/adms/representationtechnique/KIF': u'KIF',
    'http://purl.org/adms/representationtechnique/OWL': u'OWL',
    'http://purl.org/adms/representationtechnique/Prolog': u'Prolog',
    'http://purl.org/adms/representationtechnique/RDFSchema': u'RDFS',
    'http://purl.org/adms/representationtechnique/RIF': u'RIF',
    'http://purl.org/adms/representationtechnique/RelaxNG': u'Relax NG',
    'http://purl.org/adms/representationtechnique/RuleML': u'RuleML',
    'http://purl.org/adms/representationtechnique/SBVR': u'SBVR',
    'http://purl.org/adms/representationtechnique/SKOS': u'SKOS',
    'http://purl.org/adms/representationtechnique/SPARQL': u'SPARQL',
    'http://purl.org/adms/representationtechnique/SPIN': u'SPIN',
    'http://purl.org/adms/representationtechnique/SWRL': u'SWRL',
    'http://purl.org/adms/representationtechnique/Schematron': u'Schematron',
    'http://purl.org/adms/representationtechnique/TopicMaps': u'Carte thématique',
    'http://purl.org/adms/representationtechnique/UML': u'UML',
    'http://purl.org/adms/representationtechnique/WSDL': u'WSDL',
    'http://purl.org/adms/representationtechnique/WSMO': u'WSMO',
    'http://purl.org/adms/representationtechnique/XMLSchema': u'XSD',
    'http://purl.org/adms/status/Completed': u'Terminé',
    'http://purl.org/adms/status/Deprecated': u'Deprécié',
    'http://purl.org/adms/status/UnderDevelopment': u'En développement',
    'http://purl.org/adms/status/Withdrawn': u'Abandonné',
}


def add_source(cnx, name, url):
    return cnx.create_entity('SKOSSource', name=name, url=url)


def datapath(fname):
    return join(dirname(__file__), 'migration', 'data', fname)


def media_types_extentities(media_types=None):
    """Yield ExtEntity objects fetch from parsing IANA CSV files from
    http://www.iana.org/assignments/media-types/media-types.xml.

    If media_types is specified, it should be a list of domain to import.
    Otherwise all domains will be imported.
    """
    iana_uri = 'http://www.iana.org/assignments/media-types/media-types.xml'
    yield ExtEntity('ConceptScheme', iana_uri, {'title': set([u'IANA Media Types'])})
    if media_types is None:
        media_types = ('application', 'audio', 'image', 'message', 'model',
                       'multipart', 'text', 'video')
    for typename in media_types:
        with open(datapath(typename + '.csv')) as f:
            reader = ucsvreader(f, encoding='utf-8', delimiter=',', skipfirst=True)
            concepts = set([])
            for line in reader:
                fulltypename = typename + '/' + line[0]
                if fulltypename in concepts:
                    # Only consider first occurences.
                    continue
                concepts.add(fulltypename)
                yield ExtEntity('Concept', fulltypename, {'in_scheme': set([iana_uri])})
                yield ExtEntity('Label', fulltypename + '_label',
                                {'label': set([fulltypename]),
                                 'language_code': set([u'en']),
                                 'kind': set([u'preferred']),
                                 'label_of': set([fulltypename])})


def media_types_import(cnx, **kwargs):
    """Import of IANA media types concepts from CSV files."""
    import_log = SimpleImportLog('Media Types')
    if cnx.repo.system_source.dbdriver == 'postgres':
        store = massive_store.MassiveObjectStore(cnx)
    else:
        store = stores.NoHookRQLObjectStore(cnx)
    stats, (scheme, ) = import_skos_extentities(
        cnx, media_types_extentities(**kwargs), import_log, store=store,
        raise_on_error=True)
    return stats, scheme


def capitalized_eurovoc_domain(domain):
    """Return the given Eurovoc domain name with only the first letter uppercase."""
    return re.sub(r'^(\d+\s)(.+)$',
                  lambda m: u'{0}{1}'.format(m.group(1), m.group(2).lower().capitalize()),
                  domain, re.UNICODE)


def simplify_eurovoc(eurovoc_graph, new_graph):  # noqa (mccabe too complex)
    """Simplify the original Eurovoc thesaurus so that it can be used in cubicweb-skos.

    First given graph (``eurovoc_graph``) is an rdfio graph containing original Eurovoc thesaurus
    in RDF. Then, second graph (``new_graph``) will be populated with simplified triples.

    Actual simplifications are done like the this:

    * each of the 21 Eurovoc domain is used as a concept scheme with a ``dcterms:title`` property
    * each of the 127 Eurovoc micro thesaurus is used as a concept (it thus becomes a top concept
      of the corresponding domain),
    * for each micro thesaurus (which becomes a concept), its former top concepts become narrower
      concepts,
    * for each micro thesaurus, all of its concepts become concepts of the coresponding domain
      concept scheme (that is ``skos:inScheme`` property object become the domain).

    """
    reg = rdfio.RDFRegistry()
    uri = reg.normalize_uri
    uriobject = eurovoc_graph.uri
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_prefix('skos-xl', 'http://www.w3.org/2008/05/skos-xl#')
    reg.register_prefix('eu', 'http://eurovoc.europa.eu/schema#')
    reg.register_prefix('thes', 'http://purl.org/iso25964/skos-thes#')
    eurovoc_uri = u'http://eurovoc.europa.eu/100141'
    microthes2domain = dict((mt_uri, next(eurovoc_graph.objects(mt_uri, uri('eu:domain'))))
                            for mt_uri in eurovoc_graph.uris_for_type(uri('eu:MicroThesaurus')))
    for subj_uri, pred_uri, obj_uri in eurovoc_graph.triples():
        is_active = (u'http://publications.europa.eu/resource/authority/status/deprecated' not in
                     list(eurovoc_graph.objects(subj_uri, uri('thes:status'))))
        # Convert eu:Domain into a skos:ConceptScheme
        if obj_uri == uri('eu:Domain'):
            new_graph.add(uriobject(subj_uri), uriobject(uri('rdf:type')),
                          uriobject(uri('skos:ConceptScheme')))
            for title in eurovoc_graph.objects(subj_uri, uri('skos:prefLabel')):
                if title.lang == u'fr':
                    new_graph.add(uriobject(subj_uri), uriobject(uri('dcterms:title')),
                                  capitalized_eurovoc_domain(title))
                    break
        # Convert eu:MicroThesaurus into a skos:Concept
        elif obj_uri == uri('eu:MicroThesaurus'):
            new_graph.add(uriobject(subj_uri), uriobject(uri('rdf:type')),
                          uriobject(uri('skos:Concept')))
            scheme_uri = next(eurovoc_graph.objects(subj_uri, uri('eu:domain')))
            new_graph.add(uriobject(subj_uri), uriobject(uri('skos:inScheme')),
                          uriobject(scheme_uri))
        # Convert eu:ThesaurusConcept into a skos:Concept if not deprecated
        elif obj_uri == uri('eu:ThesaurusConcept'):
            new_graph.add(uriobject(subj_uri), uriobject(uri('rdf:type')),
                          uriobject(uri('skos:Concept')))
        # Replace <concept> topConceptOf <MicroThesaurus> by <concept> broader <MicroThesaurus>
        elif is_active and pred_uri == uri('skos:topConceptOf'):
            new_graph.add(uriobject(subj_uri), uriobject(uri('skos:broader')), uriobject(obj_uri))
        # Replace <concept> skos:inScheme <MicroThes> by <concept> skos:broader <MicroThes>
        # and <concept> skos:inScheme <Domain>
        elif is_active and pred_uri == uri('skos:inScheme') and obj_uri in microthes2domain:
            new_graph.add(uriobject(subj_uri), uriobject(uri('skos:inScheme')),
                          uriobject(microthes2domain[obj_uri]))
        # Keep label triples
        elif (is_active and obj_uri != eurovoc_uri and subj_uri != eurovoc_uri and
              pred_uri in (uri('skos:prefLabel'), uri('skos:altLabel'), uri('skos:hiddenLabel'))):
            new_graph.add(uriobject(subj_uri), uriobject(pred_uri), obj_uri)
        # Keep existing skos:broader relations
        elif is_active and pred_uri == uri('skos:broader'):
            new_graph.add(uriobject(subj_uri), uriobject(pred_uri), uriobject(obj_uri))
        # Keep existing concepts
        elif is_active and obj_uri == uri('skos:Concept'):
            new_graph.add(uriobject(subj_uri), uriobject(pred_uri), uriobject(obj_uri))


def eurovoc_domains(simplified_graph):
    """Return a dictionary containing all Eurovoc domains with their URIs.

    Keys are domain URIs and values are domain titles (in french).
    """
    reg = rdfio.RDFRegistry()
    uri = reg.normalize_uri
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    domains = {}
    for domain_uri in simplified_graph.uris_for_type(uri('skos:ConceptScheme')):
        domains[domain_uri] = unicode(next(simplified_graph.objects(domain_uri,
                                                                    uri('dcterms:title'))))
    return domains



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat geo-spatial schema definitions"""

from cubicweb import _
from yams.buildobjs import (
    EntityType,
    RelationDefinition,
    String,
)

from cubes.postgis.schema import Geometry


class Location(EntityType):
    """A location is used to locate objects on the Earth's surface"""
    name = String(required=True, fulltextindexed=True,
                  description=_("Name of the location (eg. 'France', 'Boston', "
                                "'President Kennedy bus stop')"))
    geometry = Geometry(
        geom_type='GEOMETRYCOLLECTION', srid=4326, coord_dimension=2,
        description=_('Geospatial indication of the location')
    )


class locate_datasets(RelationDefinition):
    subject = 'Location'
    object = 'Dataset'
    cardinality = '**'
    description = _('relate a location to one or more datasets')



"""cubicweb-datacat unit tests for server objects"""

import tempfile
import urllib2

from logilab.mtconverter import xml_escape

from cubicweb.devtools import (
    PostgresApptestConfiguration,
    startpgcluster,
    stoppgcluster,
    testlib,
)

from cubes.datacat import cwsource_pull_data
from utils import (
    create_file,
    mediatypes_scheme,
)


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


def transformation_sequence(cnx, script):
    tseq = cnx.create_entity('TransformationSequence')
    cnx.create_entity('TransformationStep', index=0,
                      step_script=script, in_sequence=tseq)
    return tseq


class BaseResourceFeedTC(testlib.CubicWebTC):

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', title=u'Test dataset', description=u'A dataset',
                                   in_catalog=cat)
            cnx.commit()
            self.dataset_eid = ds.eid


class ResourceFeedParserTC(BaseResourceFeedTC):

    def pull_data(self, cwsource, **kwargs):
        return cwsource_pull_data(self.repo, cwsource.eid, **kwargs)

    def test_cwsource_without_resourcefeed(self):
        with self.admin_access.cnx() as cnx:
            source = cnx.create_entity(
                'CWSource', parser=u'datacat.resourcefeed-parser',
                type=u'datafeed', name=u'test', url=u'does not matter')
            cnx.commit()
        with self.assertRaisesRegexp(Exception, 'has no ResourceFeed attached'):
            self.pull_data(source, raise_on_error=True)

    def test_base(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
        cwsource = resourcefeed.resource_feed_source[0]
        dist = resourcefeed.resourcefeed_distribution[0]
        stats = self.pull_data(cwsource)
        self.assertEqual(len(stats['created']), 1)
        feid = stats['created'].pop()
        with self.admin_access.cnx() as cnx:
            importlog = cnx.find('CWDataImport').one()
            logs = importlog.log
        self.assertIn('fetched data from {0}'.format(url), logs)
        self.assertIn('data import for {0} completed'.format(url), logs)
        self.assertIn('adding file #{0} to distribution #{1}'.format(
            feid, dist.eid), logs)
        self.assertIn(xml_escape('adding file #{0} to distribution').format(
            feid), logs)
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.entity_from_eid(feid)
            self.assertEqual(f.data_name, 'resource.dat')
            self.assertEqual(f.data_format, 'text/csv')
            self.assertEqual(list(f.file_distribution), [dist])
        # second pull, no update of sha1
        stats = self.pull_data(cwsource)
        self.assertEqual(len(stats['created']), 0)

    def test_with_processes(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            # Validation script.
            vscript_eid = cnx.create_entity(
                'ValidationScript', name=u'validation script',
                media_type=mediatype).eid
            create_file(cnx, 'pass', reverse_implemented_by=vscript_eid)
            # Transformation sequence.
            tscript = cnx.create_entity(
                'TransformationScript',
                name=u'transformation script',
                media_type=mediatype)
            create_file(cnx, open(self.datapath('cat.py')).read(),
                        data_name=u'cat.py',
                        reverse_implemented_by=tscript)
            # Create resource feed.
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid,
                validation_script=vscript_eid,
                transformation_script=tscript)
            cnx.commit()
        cwsource = resourcefeed.resource_feed_source[0]
        stats = self.pull_data(cwsource)
        assert len(stats['created']) == 1
        with self.admin_access.cnx() as cnx:
            importlog = cnx.find('CWDataImport').one()
            logs = importlog.log
        self.assertIn('created validation process', logs)
        self.assertIn('created transformation process', logs)
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.find('DataTransformationProcess',
                            process_for_resourcefeed=resourcefeed.eid)
            process = rset.one()
            self.assertEqual(process.in_state[0].name,
                             'wfs_dataprocess_completed')
            # There should be one result.
            output = cnx.find('File', produced_by=process).one()
            self.assertEqual([r.eid for r in output.file_distribution[0].of_dataset],
                             [self.dataset_eid])

    def test_file_update(self):
        """Update a file between two datafeed pulls"""
        with tempfile.NamedTemporaryFile() as tmpf:
            tmpf.write('coucou')
            tmpf.flush()
            with self.admin_access.repo_cnx() as cnx:
                mtype, = mediatypes_scheme(cnx, u'whatever')
                vscript_eid = cnx.create_entity(
                    'ValidationScript',
                    name=u'validation script',
                    media_type=mtype).eid
                create_file(cnx, 'pass',
                            reverse_implemented_by=vscript_eid)
                # Transformation script and process.
                tscript_eid = cnx.create_entity(
                    'TransformationScript',
                    name=u'transformation script',
                    media_type=mtype).eid
                create_file(cnx, open(self.datapath('reverse.py')).read(),
                            data_name=u'reverse',
                            reverse_implemented_by=tscript_eid)
                resourcefeed = cnx.create_entity(
                    'ResourceFeed', url=u'file://' + tmpf.name,
                    media_type=mtype,
                    validation_script=vscript_eid,
                    transformation_script=tscript_eid,
                    # Setting a "large" versions_to_keep value to prevent
                    # deletion of replaced files.
                    versions_to_keep=10,
                    resource_feed_of=self.dataset_eid)
                cnx.commit()
            cwsource = resourcefeed.resource_feed_source[0]
            stats = self.pull_data(cwsource)
            self.assertEqual(len(stats['created']), 1, stats)
            feid = stats['created'].pop()
            expected = {'content': 'uocuoc\n', 'nvalidated': 1, 'nproduced': 1}
            self._check_datafeed_output(feid, vscript_eid,
                                        tscript_eid, expected)
            # Change input file.
            tmpf.write('\nau revoir')
            tmpf.flush()
            stats = self.pull_data(cwsource)
            self.assertEqual(len(stats['created']), 1)
            feid_ = stats['created'].pop()
            # second pull: change input
            expected = {'content': 'riover ua\nuocuoc\n',
                        'nvalidated': 2, 'nproduced': 2}
            self._check_datafeed_output(feid_, vscript_eid,
                                        tscript_eid, expected)
            # Pull one more time, without changing the source.
            # third pull: no change
            stats = self.pull_data(cwsource)
            for k, v in stats.iteritems():
                self.assertFalse(v, '%s: %r' % (k, v))
            # `expected` has not changed.
            self._check_datafeed_output(feid_, vscript_eid,
                                        tscript_eid, expected)

    def _check_datafeed_output(self, feid, vscript_eid, tscript_eid, expected):
        with self.admin_access.repo_cnx() as cnx:
            output = cnx.execute(
                'File X WHERE X produced_by TP, TP process_input_file F, F eid %(f)s',
                {'f': feid}).one()
            if 'content' in expected:
                self.assertEqual(output.read(), expected['content'])
            if 'nvalidated' in expected:
                rset = cnx.execute(
                    'Any X WHERE X validated_by VP, VP validation_script S, S eid %(s)s',
                    {'s': vscript_eid})
                self.assertEqual(len(rset), expected['nvalidated'], rset)
            if 'nproduced' in expected:
                nproduced = expected['nproduced']
                rset = cnx.execute(
                    'Any X WHERE X produced_by TP, TP transformation_sequence SEQ,'
                    ' STEP in_sequence SEQ, STEP step_script S, S eid %(s)s',
                    {'s': tscript_eid})
                self.assertEqual(len(rset), nproduced, rset)

    def _check_value_error(self, cwsource):
        with self.assertRaises(ValueError) as cm:
            self.pull_data(cwsource)
        self.assertIn('MIME types of resource feeds attached',
                      str(cm.exception))

    def _check_transformations(self, cnx, stats):
        # Check both data processes have completed.
        rset = cnx.execute(
            'DataTransformationProcess X WHERE X in_state ST,'
            ' ST name "wfs_dataprocess_completed"')
        self.assertEqual(len(rset), 2)
        # Check resources and produced files.
        self.assertEqual(
            cnx.execute('Any COUNT(X) WHERE X file_distribution D')[0][0], 2)
        self.assertEqual(len(stats['created']), 1, stats)
        feid = stats['created'].pop()
        rset = cnx.execute('File X WHERE X produced_by TP, TP process_input_file F, F eid %(f)s',
                           {'f': feid})
        self.assertEqual(len(rset), 2, rset)

    def test_ftp(self):
        """Check that datafeed monkeypatch for retrieve_url handles ftp protocol"""
        url = u'ftp://user:pwd@does.not.exists'
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            cwsource = resourcefeed.resource_feed_source[0]
            dfsource = self.repo.sources_by_eid[cwsource.eid]
            parser = dfsource._get_parser(cnx)
            # Just check we get an urllib2 error, meaning that we went through
            # opening the url.
            self.assertRaises(urllib2.URLError, parser.retrieve_url, url)


class GargageRessourceFeedProcessesTC(BaseResourceFeedTC):

    def test_cleanup(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            vscript_eid = cnx.create_entity(
                'ValidationScript', name=u'validation script',
                media_type=mediatype)
            create_file(cnx, 'console.log("bad language ;)")',
                        reverse_implemented_by=vscript_eid)
            resourcefeed1 = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid,
                validation_script=vscript_eid,
                processes_to_keep=2)
            resourcefeed2 = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid,
                validation_script=vscript_eid,
                processes_to_keep=1)
            datafile = create_file(cnx, '1,2,3')
            cnx.commit()
            with cnx.security_enabled(write=False):
                processes1 = []
                for _ in range(3):
                    processes1.append(
                        cnx.create_entity('DataValidationProcess',
                                          process_for_resourcefeed=resourcefeed1,
                                          validation_script=vscript_eid,
                                          process_input_file=datafile))
                    cnx.commit()
                processes2 = []
                for _ in range(3):
                    processes2.append(
                        cnx.create_entity('DataValidationProcess',
                                          process_for_resourcefeed=resourcefeed2,
                                          validation_script=vscript_eid,
                                          process_input_file=datafile))
                    cnx.commit()

            def process_logfile(p):
                p.cw_clear_all_caches()
                return p.process_stderr[0].eid if p.process_stderr else None

            logfiles1 = map(process_logfile, processes1)
            logfiles2 = map(process_logfile, processes2)

            def process_log_files(eid):
                """Log files of processes of a resource feed with `eid`."""
                rset = cnx.execute(
                    'Any F WHERE P process_stderr F, P process_for_resourcefeed R,'
                    ' R eid %(eid)s', {'eid': eid})
                return [x for x, in rset.rows]

            with self.assertLogs(level='INFO') as cm:
                cnx.call_service('datacat.clean-resourcefeed-process-logs',
                                 eid=resourcefeed1.eid)
            self.assertEqual(len(cm.output), 1)
            self.assertIn('deleted process log files "%s"' % logfiles1[0],
                          str(cm.output[0]))

            self.assertCountEqual(process_log_files(resourcefeed1.eid),
                                  logfiles1[-2:])
            self.assertCountEqual(process_log_files(resourcefeed2.eid),
                                  logfiles2)

            with cnx.security_enabled(write=False):
                newprocess = cnx.create_entity(
                    'DataValidationProcess',
                    process_for_resourcefeed=resourcefeed1,
                    validation_script=vscript_eid,
                    process_input_file=datafile)
                cnx.commit()
                newprocess.cw_clear_all_caches()
                logfiles1.append(newprocess.process_stderr[0].eid)

            with self.assertLogs(level='INFO') as cm:
                cnx.call_service('datacat.clean-resourcefeed-process-logs')
            self.assertEqual(len(cm.output), 2)
            self.assertCountEqual(process_log_files(resourcefeed2.eid),
                                  logfiles2[-1:])
            self.assertCountEqual(process_log_files(resourcefeed1.eid),
                                  logfiles1[-2:])


if __name__ == '__main__':
    import unittest
    unittest.main()



# New entity Catalog
add_entity_type('DataCatalog')
# More properties on Dataset and Distribution
add_attribute('Dataset', 'issued')
add_attribute('Dataset', 'modified')
rql('SET X title T WHERE X is Dataset, X title NULL, X identifier T')
commit()
sync_schema_props_perms('Dataset')
add_attribute('Distribution', 'title')
add_attribute('Distribution', 'issued')
add_attribute('Distribution', 'modified')
add_attribute('Distribution', 'download_url')
sync_schema_props_perms('Distribution')
# Rename relation between Dataset and Distribution
# Also invert subject and object
add_relation_type('of_dataset')
rql('Set Y of_dataset X WHERE X dataset_distribution Y, X is Dataset, Y is Distribution')
commit()
drop_relation_type('dataset_distribution')

for etype in ('DataTransformationProcess', 'DataValidationProcess'):
    sync_schema_props_perms(etype)
    drop_attribute(etype, 'description')
    drop_attribute(etype, 'description_format')

rename_attribute('ResourceFeed', 'uri', 'url')
add_attribute('ResourceFeed', 'title')
sync_schema_props_perms('ResourceFeed')

add_relation_type('resourcefeed_distribution')
for resourcefeed in find_entities('ResourceFeed'):
    create_entity('Distribution', format=resourcefeed.data_format,
                  reverse_resourcefeed_distribution=resourcefeed,
                  of_dataset=resourcefeed.resource_feed_of[0])
    commit()

add_relation_type('file_distribution')
for f in rql('Any F WHERE EXISTS(F resource_of D)').entities():
    sourceinfo = f.cw_metainformation()['source']
    if sourceinfo['type'] == 'datafeed':
        # Relate file produced by a ResourceFeed to the associated
        # distribution.
        uri = sourceinfo['uri']
        source_eid = repo.sources_by_uri[uri].eid
        rql('SET F file_distribution D WHERE'
            ' RS resource_feed_source S, RS resourcefeed_distribution D,'
            ' S eid %(seid)s, F eid %(feid)s',
            {'seid': source_eid, 'feid': f.eid})
    else:
        # If the file is not from datafeed, create a distribution of it.
        create_entity('Distribution', title=f.title, format=f.data_format,
                      description=f.description,
                      description_format=f.description_format,
                      reverse_file_distribution=f, of_dataset=f.resource_of)
    commit()

drop_relation_type('resource_of')
sync_schema_props_perms('File', syncprops=False)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat geo-spatial views/forms/actions/components"""

import io

# XXX: find a way to load Postgis WKB from osgeo.ogr and thus avoid to depends
# on shapely in addition to osgeo
from shapely import wkb

from logilab.mtconverter import xml_escape
from rql import nodes

from cubicweb import _
from cubicweb.predicates import (
    adaptable,
    is_instance,
)
from cubicweb.utils import (
    json_dumps,
    JSString,
    make_uid,
)
from cubicweb import tags
from cubicweb.web import (
    component,
    facet as cwfacet,
    formwidgets,
    htmlwidgets,
    ProcessFormError,
)
from cubicweb.web.views import (
    primary,
    uicfg,
)

from cubes.datacat.geoutils import merge_in_geojson_collection
from cubes.leaflet import views as leafletviews


affk = uicfg.autoform_field_kwargs
afs = uicfg.autoform_section
pvs = uicfg.primaryview_section


#
# Leaflet maps and views
#
#

class LeafletJSONPropertiesMap(leafletviews.LeafletMultiPolygonValues):
    """Leaflet map for plotting GeoJSON data and associated properties.

    Unlike parent class, this one computes initial zoom in order to fit data at loading time.

    Should be used in a view.
    """
    # See https://www.cubicweb.org/ticket/5630012
    # TODO: drop this class when ticket is done

    default_settings = {
        'legend_visibility': False,
        'info_visibility': True,
        'info_position': 'topright',
    }

    js_plotter = 'cw.cubes.datacat.plotMapProperties'


class LeafletGeoJSONEntityView(leafletviews.AbstractLeafletView):
    """Plot IGeoJSON adaptable entities in a Leaflet map."""

    __regid__ = 'leaflet-geojson-entities'
    __select__ = leafletviews.AbstractLeafletView.__select__ & adaptable('IGeoJSON')

    plotclass = LeafletJSONPropertiesMap

    def call(self, *args, **kwargs):
        self._cw.add_js('cubes.datacat.js')
        super(LeafletGeoJSONEntityView, self).call(*args, **kwargs)

    def build_markers(self):
        """Build the GeoJSON data from entities in the view."""
        geojsons = (entity.cw_adapt_to('IGeoJSON').as_geojson()
                    for entity in self.cw_rset.entities())
        data = merge_in_geojson_collection(geojsons)
        return JSString(json_dumps(data))


class DrawableLeafletJsonMap(leafletviews.LeafletMultiPolygon):
    """Leaflet map for plotting GeoJSON data and allowing draws

    Unlike parent class, this one computes initial zoom in order to fit data at loading time.

    Should be used in a view.
    """

    default_settings = leafletviews.LeafletMultiPolygon.default_settings.copy()
    default_settings.update({
        'draw_settings': {
            'polyline': {},
            'polygon': {},
            'rectangle': {},
            'circle': {},
            'marker': {},
        },
        'edit_settings': {
            'edit': {},
            'remove': {},
        }
    })
    js_plotter = 'cw.cubes.datacat.plotDrawableMap'

    def render(self, req, datasource, use_cdn=True):
        """Renders the map."""
        # Need to call the parent method first because Leaflet.Draw needs Leaflet which is loaded
        # in parent view
        div_holder = super(DrawableLeafletJsonMap, self).render(req, datasource, use_cdn=use_cdn)
        req.add_js('cubes.datacat.js')
        if use_cdn:
            req.add_css(
                'https://cdnjs.cloudflare.com/ajax/libs/leaflet.draw/0.2.3/leaflet.draw.css',
                localfile=False)
            req.add_js(
                'https://cdnjs.cloudflare.com/ajax/libs/leaflet.draw/0.2.3/leaflet.draw.js',
                localfile=False)
        else:
            req.add_css('leaflet.draw.css')
            req.add_js('leaflet.draw.js')
        return div_holder


class DrawableLeafletJsonView(leafletviews.GeoJsonView):
    """View for plotting GeoJSON data in an drawable Leaflet map."""

    __regid__ = 'leaflet-geojson-draw'

    plotclass = DrawableLeafletJsonMap


class GeoJSONEntityCtxComponent(component.EntityCtxComponent):
    """Display geospatial data in a Leaflet view."""
    __regid__ = 'datacat.geojson-entity-component'
    __select__ = component.EntityCtxComponent.__select__ & adaptable('IGeoJSON')
    title = _('Location')
    context = 'navcontentbottom'

    def render_body(self, w):
        self._cw.view('leaflet-geojson-entities', rset=self.entity.as_rset(), w=w)


class FacetGeoJSONWidget(htmlwidgets.HTMLWidget):
    """Map widget to be used in a facet."""

    default_settings = {
        'width': '300px',
        'height': '300px',
        'drawSettings': {},
        'editSettings': {},
    }

    def __init__(self, facet, rset, leaflet_vid='leaflet-geojson-draw', custom_settings=None):
        self.facet = facet
        self.rset = rset
        self.leaflet_vid = leaflet_vid
        self.geojsonpolygon = facet._cw.form.get(xml_escape(facet.__regid__) + 'Input')
        # Map settings
        self.settings = self.default_settings.copy()
        self.settings.update(custom_settings or {})

    @property
    def height(self):
        """Height of the widget"""
        return 4.5

    def _render(self):
        """Render the widget as a Leaflet Map."""
        w = self.w
        facet = self.facet
        cnx = facet._cw
        # Useful vars
        facettitle = xml_escape(facet.title)
        facetname = xml_escape(facet.__regid__)
        inputname = facetname + 'Input'
        divid = make_uid('mapFacet')
        facetid = make_uid(facet.__regid__)
        self.settings.update({'divid': divid, 'inputName': inputname})
        w(u'<div id="{}" class="facet">\n'.format(facetid))  # Facet opening <div>
        # Facet title
        cssclass = 'facetTitle' if not facet.allow_hide else 'facetTitle hideFacetBody'
        w(u'<div class="{}" cubicweb:facetName="{}">{}</div>\n'.format(
            cssclass, facetname, facettitle))
        # Facet body
        cssclass = 'facetBody' if facet.start_unfolded else 'facetBody hidden'
        w(u'<div class="{}">\n'.format(cssclass))  # Facet body opening <div>
        cnx.view(self.leaflet_vid, rset=self.rset, w=w, custom_settings=self.settings)
        w(u'<input type="hidden" name="{}" value="{}" />\n'.format(
            inputname, self.geojsonpolygon or u''))
        w(u'</div>\n')  # Facet body closing </div>
        w(u'</div>\n')  # Facet closing </div>


#
# GeoJSON facet
#

class RelationExtentFacet(cwfacet.AbstractFacet):
    """Base facet to filter an entity list according to a geospatial intersection criteria on other
    entities to which they are related.

    This facet displays a geographical map with the extent (that is, the convex hull) of all the
    related entities. We can draw polygons in this map to retain only entities who intersect the
    drawn polygons.

    Attributes to be declared for this facet are the same as for
    :class:`cwfacet.RelationAttributeFacet`.

    For example, let us consider a schema where offices are related to cities, and cities have a
    geometry attribute:

    .. code-block: python

        class City(EntityType):
            name = String(description='City name')
            geometry = Geometry(description='City geospatial location',
                                geom_type='MULTIPOLYGON', srid=4326, coord_dimension=2)

        class Office(EntityType):
            name = String(description='Office name')
            in_city = SubjectRelation('City', cardinality='1?')

    With this schema, we can define a facet to filter offices according to their city geometry:

    .. code-block: python

        class OfficeCityGeometryFacet(RelationExtentFacet):
            __regid__ = 'office_city_geom_facet'
            __select__ = is_instance('Office')
            rtype = 'in_city'
            role = 'subject'
            target_attr = 'geometry'

    Then in the web interface, when we are displaying a list of offices, we see a map among the
    facets. This map shows us the extent of all cities related to offices in the list. We can draw
    polygons on this map: this will select only cities that intersect the polygons, and consequently
    only offices that are related to the selected cities will remain.
    """

    target_type = None
    target_attr_type = 'Geometry'
    title = property(cwfacet.rtype_facet_title)

    default_settings = {
        'drawSettings': {
            'polyline': False,
            'marker': False,
            'circle': False,
        },
    }

    @property
    def wdgclass(self):
        """The widget's class that will be used to display the facet."""
        return FacetGeoJSONWidget

    def get_widget(self):
        """Return the widget to be used to display the convex hull, or ``None`` if there is no
        convex hull.
        """
        rset = self._geojson_convex_hull_rset()
        if rset:
            return self.wdgclass(self, rset, custom_settings=self.default_settings)
        return None

    def add_rql_restrictions(self):
        """Add an ``ST_INTERSECTS`` restriction with geometries drawn in the widget.

        Namely, add an intersection filter (``ST_INTERSECTS(G, <drawn_polygons>)``) to select
        entities that intersect the drawn geometries.
        """
        select = self.select
        # Geometry RQL node for drawn polygons
        # (eg. ST_COLLECTIONEXTRACT(ST_GEOMFROMGEOJSON(<drawngeojsonstr>), 3))
        # Note that we must extract Polygons since Leaflet widget returns GeometryCollection
        geojsonstr = self._cw.form.get(xml_escape(self.__regid__) + 'Input')
        if not geojsonstr:
            return
        select.set_distinct(True)
        drawnvar = nodes.Constant(geojsonstr, 'String')
        geomfromjsonf = nodes.Function('ST_GEOMFROMGEOJSON')
        geomfromjsonf.append(drawnvar)
        drawnnode = nodes.Function('ST_COLLECTIONEXTRACT')
        drawnnode.append(geomfromjsonf)
        drawnnode.append(nodes.Constant(3, 'Int'))
        # Target geometry RQL node for entities (eg. G WHERE E geometry G, E related X)
        targetvar = self._restriction_geometry_variable()
        # Target geometry type (eg. MultiPolygon)
        target_geomtype = self._target_geom_type()
        # List all possible intersections: Polygon/Point, Polygon/Line, Polygon/Polygon
        intersectnodes = []
        for entitynode in self._intersect_operand(targetvar, target_geomtype):
            intersectf = nodes.Function('ST_INTERSECTS')
            intersectf.append(drawnnode)
            intersectf.append(entitynode)
            comp = nodes.Comparison('=', intersectf)
            comp.append(nodes.Constant('True', 'Boolean'))
            intersectnodes.append(comp)
        # OR all intersections and insert in HAVING clause
        for i, intersectnode in enumerate(intersectnodes):
            if i == 0:
                havingnode = intersectnode
            else:
                havingnode = nodes.Or(havingnode, intersectnode)
        having = select.having
        if not having:
            select.set_having([havingnode])
        else:
            select.replace(having[0], nodes.And(havingnode, having[0]))

    def _geojson_convex_hull_rset(self):
        """Return a one-cell result set: the convex hull of all selected entities in GeoJSON."""
        select = self.select
        select.save_state()
        try:
            cwfacet.cleanup_select(select, self.filtered_variable)
            geomvar = self._convex_hull_geometry_variable()
            select_collectf = nodes.Function('ST_COLLECT')
            select_collectf.append(nodes.VariableRef(geomvar))
            # Restrict to non-empty ST_COLLECT(<geomvar>) because otherwise the
            # result set would not be empty in case no entity has a location
            # with a non-empty geometry.
            # (See http://postgis.net/docs/ST_Collect.html ST_COLLECT is an
            # aggregate function.)
            having_collectf = nodes.Function('ST_COLLECT')
            having_collectf.append(nodes.VariableRef(geomvar))
            comp = nodes.Comparison('!=')
            comp.append(having_collectf)
            # Comparing to NULL does not work properly so compare with empty
            # string.
            comp.append(nodes.Constant('', 'String'))
            select.set_having([comp])
            convexhullf = nodes.Function('ST_CONVEXHULL')
            convexhullf.append(select_collectf)
            asgeojsonf = nodes.Function('ST_ASGEOJSON')
            asgeojsonf.append(convexhullf)
            select.add_selected(asgeojsonf)
            try:
                return self.rqlexec(select.as_string(), self.cw_rset.args)
            except Exception:
                self.exception('Error while parsing values for {} (rql: "{}")'.format(
                    self, select.as_string()))
                return ()
        finally:
            select.recover()
        return None

    def _convex_hull_geometry_variable(self):
        """Return a RQL variable representing the Geometry attribute to be used for computing convex
        hull."""
        var = cwfacet.insert_attr_select_relation(self.select, self.filtered_variable, self.rtype,
                                                  self.role, self.target_attr, sortasc=None)
        if self.target_type is not None:
            self.select.add_type_restriction(var, self.target_type)
        # var represent target entity, we want geometry attribute on the target entity
        for varref in self.select.selection[:]:
            if varref.variable != var:
                geomvar = varref.variable
            self.select.remove_selected(varref)
        return geomvar

    def _restriction_geometry_variable(self):
        """Return a RQL variable representing the target Geometry attribute to be used in the added
        RQL restriction."""
        filtered_variable = self.filtered_variable
        # Add X rtype E (or E rtype X) restriction (E: related entity with geometry attribute)
        relentityvar = self.select.make_variable()
        if self.role == 'subject':
            rel = nodes.make_relation(filtered_variable, self.rtype, (relentityvar,),
                                      nodes.VariableRef)
        else:
            rel = nodes.make_relation(relentityvar, self.rtype, (filtered_variable,),
                                      nodes.VariableRef)
        self.select.add_restriction(rel)
        # Add E target_attr G restriction to query (G: target Geometry attribute)
        geomvar = self.select.make_variable()
        geomrel = nodes.make_relation(relentityvar, self.target_attr, (geomvar,),
                                      nodes.VariableRef)
        self.select.add_restriction(geomrel)
        return geomvar

    def _target_geom_type(self):
        """Return the geometry type of the target attribute."""
        rschema = self._cw.vreg.schema.rschema(self.target_attr)
        rdef = rschema.rdefs.itervalues().next()  # XXX: if multiple attr with same name exists ?
        return rdef.geom_type

    def _intersect_operand(self, var, geom_type):
        """Yield in turn each RQL node to be used as an operand in the ``ST_INTERSECTS`` function.

        For example, if ``G ``is an RQL variable representing a ``Geometry`` attribute with
        ``geom_type`` ``MultiPolygon``, then it will yield only G.

        Things get complicated when ``geom_type`` is ``GeometryCollection``. Since Postgis
        ``ST_INTERSECTS`` function does not work with geometry collections, we must extract each
        geometry type (``1: Point, 2: LineString, 3: Polygon``) and yield each one in turn.
        """
        # ST_Intersect refuses GeometryCollection (http://postgis.net/docs/ST_Intersects.html)
        if 'GEOMETRYCOLLECTION' not in geom_type.upper():
            yield var
        else:
            for i in range(1, 4):  # http://postgis.net/docs/ST_CollectionExtract.html
                extractf = nodes.Function('ST_COLLECTIONEXTRACT')
                extractf.append(var)
                extractf.append(nodes.Constant(i, 'Int'))
                yield extractf


class ExtentFacet(RelationExtentFacet):
    """Base facet to filter an entity list according to a geospatial intersection criteria.

    This facet displays a geographical map with the extent (that is, the convex hull) of all the
    entities. We can draw polygons in this map to retains only entities which intersect the drawn
    polygons.

    Attributes to be declared for this facet are the same as for :class:`cwfacet.AttributeFacet`.

    For example, let us consider a schema where offices have a geometry attribute:

    .. code-block: python

        class Office(EntityType):
            name = String(description='Office name')
            geometry = Geometry(description='Office geospatial shape',
                                geom_type='MULTIPOLYGON', srid=4326, coord_dimension=2)

    With this schema, we can define a facet to filter offices according to their geometries:

    .. code-block: python

        class OfficeGeometryFacet(ExtentFacet):
            __regid__ = 'office_geom_facet'
            __select__ = is_instance('Office')
            rtype = 'geometry'

    Then in the web interface, when we are displaying a list of offices, we see a map among the
    facets. This map shows us the extent of all offices in our list. We can draw polygons on this
    map: this will select only offices that intersect the drawn polygons.
    """

    role = 'subject'

    def _convex_hull_geometry_variable(self):
        """Return a RQL variable representing the Geometry attribute to be used for computing convex
        hull."""
        return cwfacet._add_rtype_relation(self.select, self.filtered_variable, self.rtype,
                                           self.role)[0]

    def _restriction_geometry_variable(self):
        """Return a RQL variable representing the target Geometry attribute to be used in the added
        RQL restriction."""
        filtered_variable = self.filtered_variable
        # Add X rtype G restriction (G: target Geometry attribute)
        geomvar = self.select.make_variable()
        geomrel = nodes.make_relation(filtered_variable, self.rtype, (geomvar,), nodes.VariableRef)
        self.select.add_restriction(geomrel)
        return geomvar

    def _target_geom_type(self):
        """Return the geometry type of the target attribute."""
        rschema = self._cw.vreg.schema.rschema(self.rtype)
        rdef = rschema.rdefs.itervalues().next()  # XXX: if multiple attr with same name exists ?
        return rdef.geom_type


class DatasetGeometryFacet(RelationExtentFacet):
    __regid__ = 'datacat.dataset_extent_facet'
    __select__ = is_instance('Dataset')
    rtype = 'locate_datasets'
    role = 'object'
    target_attr = 'geometry'


class LocationGeometryFacet(ExtentFacet):
    __regid__ = 'datacat.location_extent_facet'
    __select__ = is_instance('Location')
    rtype = 'geometry'


#
# Location
#


class LongitudeLatitudeWidget(formwidgets.TextInput):
    """Form widget handling POINT(x y) geometry attribute as longitude and
    latitude.
    """

    def _render(self, form, field, renderer):
        ustring = io.StringIO()
        w = ustring.write
        value = self.values(form, field)[0]
        longitude = 0
        latitude = 0
        if value:
            value = wkb.loads(value.decode('hex'))
            if not len(value) == 1:
                raise ValueError('cannot render {} in form widget'.format(value))
            point = value[0]
            longitude = point.x
            latitude = point.y

        for label, value in [(u'longitude', longitude),
                             (u'latitude', latitude)]:
            w(u'<div class="form-group">')
            w(tags.label(label))
            w(tags.input(type='text', value=value,
                         name=field.input_name(form, label)))
            w(u'</div>')

        return ustring.getvalue()

    def process_field_data(self, form, field):
        posted = form._cw.form
        longitude = posted.get(field.input_name(form, 'longitude')).strip() or None
        latitude = posted.get(field.input_name(form, 'latitude')).strip() or None
        msg = form._cw._(
            'both latitude and longitude must be specified on unspecified')
        if latitude is None:
            if longitude is not None:
                raise ProcessFormError(msg)
            return None
        if longitude is None:
            if latitude is not None:
                raise ProcessFormError(msg)
            return None
        return u'GEOMETRYCOLLECTION (POINT(%s %s))' % (longitude, latitude), 4326


# Hide geom by default (binary content)
pvs.tag_attribute(('Location', 'geometry'), 'hidden')
affk.tag_attribute(('Location', 'geometry'),
                   {'widget': LongitudeLatitudeWidget})


class LocationPrimaryView(primary.PrimaryView):
    """Override primary view for Location entities."""
    __select__ = primary.PrimaryView.__select__ & is_instance('Location')

    def render_entity_attributes(self, entity):
        super(LocationPrimaryView, self).render_entity_attributes(entity)
        self.wview('leaflet-geojson-entities', rset=entity.as_rset())
